use std::collections::HashMap;
use std::iter::Peekable;

use roxmltree::{Children, Node};

use crate::law::LawBuilder;
use crate::paragraph::parser::{
    Absatz, AbsatzAbs, Content, Fzinhalt, Kzinhalt, Liste, Table, Ueberschrift,
};

#[derive(Debug, PartialEq, Default)]
pub(crate) struct Abschnitt {
    metadata: HashMap<String, String>,
    pub(crate) cont: bool,
}

impl Abschnitt {
    pub(crate) fn parse(n: Node, builder: &mut LawBuilder) -> Abschnitt {
        assert!(n.tag_name().name() == "abschnitt");

        let mut ret = Abschnitt::default();
        let mut c = n.children().peekable();

        Self::skip_static_fields(&mut c);

        ret.handle_metadata(&mut c, builder);

        if !ret.handle_headers(&mut c, builder) {
            return ret;
        }

        let mut absatze = Vec::new();
        let absatz = AbsatzAbs::parse(c.next().expect("We need at least one 'Absatz'"));
        let par_id = absatz
            .gldsym
            .clone()
            .expect("First 'Absatz' needs to have § id");

        // If there's a  "liste" after an "absatz", the "liste" should be part of the "absatz"
        if let Some(child) = c.peek() {
            if Liste::test(child) {
                let liste = Liste::parse(c.next().unwrap());
                let mut to_add = vec![Content::Text(absatz.content), liste.get_content()];
                if let Some(subchild) = c.peek() {
                    if Absatz::test_with_typ(subchild, "satz") {
                        // After a 'liste' there can be a '<absatz typ="satz"' which should be part of the list
                        // (e.g. 85 StGB)
                        to_add.push(Content::Text(Absatz::parse(c.next().unwrap()).content));
                    }
                }
                absatze.push(Content::List(to_add));
            } else if Table::test(child) {
                // If there's a  "table" after an "absatz", the "table" should be part of the "absatz"
                let table = Table::parse(c.next().unwrap());
                if let Some(child) = c.peek() {
                    if Absatz::test_with_typ(child, "erltext") {
                        let after_absatz = Absatz::parse(c.next().unwrap());
                        absatze.push(Content::List(vec![
                            Content::Text(absatz.content),
                            Content::List(table.get_list()),
                            Content::Text(after_absatz.content),
                        ]));
                    } else {
                        absatze.push(Content::List(vec![
                            Content::Text(absatz.content),
                            Content::List(table.get_list()),
                        ]));
                    }
                }
            } else if Absatz::test_with_typ(child, "satz") {
                // After a 'liste' there can be a '<absatz typ="satz"' which should be part of the list
                // (e.g. 1209 ABGB)
                absatze.push(Content::List(vec![
                    Content::Text(absatz.content.clone()),
                    Content::Text(Absatz::parse(c.next().unwrap()).content),
                ]));
            } else {
                absatze.push(Content::Text(absatz.content.clone()));
            }
        } else {
            absatze.push(Content::Text(absatz.content.clone()));
        }

        //There can be as many 'Absätze' as our lovely lawsetter wants
        while let Some(child) = c.peek() {
            if AbsatzAbs::test(child) {
                let abs = AbsatzAbs::parse(c.next().unwrap());

                // If there's a  "liste" after an "absatz", the "liste" should be part of the "absatz"
                if let Some(child) = c.peek() {
                    if Liste::test(child) {
                        let liste = Liste::parse(c.next().unwrap());
                        let mut to_add = vec![Content::Text(abs.content), liste.get_content()];
                        if let Some(subchild) = c.peek() {
                            if Absatz::test_with_typ(subchild, "satz") {
                                // After a 'liste' there can be a '<absatz typ="satz"' which should be part of the list
                                // (e.g. 85 StGB)
                                to_add
                                    .push(Content::Text(Absatz::parse(c.next().unwrap()).content));
                            }
                        }
                        absatze.push(Content::List(to_add));
                    } else {
                        let mut content = abs.content;
                        while let Some(subchild) = c.peek() {
                            if Absatz::test_with_typ(subchild, "erltext") {
                                content += &Absatz::parse(c.next().unwrap()).content;
                            } else {
                                break;
                            }
                        }
                        absatze.push(Content::Text(content));
                    }
                } else {
                    absatze.push(Content::Text(abs.content));
                }
                continue;
            }
            break;
        }

        if absatze.len() == 1 {
            builder.new_par(par_id, absatze[0].clone());
        } else {
            let mut contents = Vec::new();
            for a in &absatze {
                contents.push(a.clone());
            }
            builder.new_par(par_id, Content::Item(contents));
        }

        // Skip all UeberschriftTitle and Absatz
        while let Some(child) = c.peek() {
            if Ueberschrift::test(child, "titel") {
                c.next();
                continue;
            }
            if Absatz::test_with_typ(child, "erltext") {
                c.next();
                continue;
            }
            break;
        }

        assert_eq!(c.next(), None);

        ret.cont = true;
        ret
    }

    // There are paragraph-specific meta-data at the top of each xml file. We parse those. When we
    // encounter the title "Text" the real content starts, we stop parsing meta data.
    fn handle_metadata(&mut self, c: &mut Peekable<Children>, builder: &mut LawBuilder) {
        loop {
            let key = Ueberschrift::parse(c.next().unwrap(), "titel").content;

            // We are done with meta-data parsing
            if key == "Text" {
                break;
            }

            let absatz = Absatz::parse(
                c.next()
                    .expect("Expected absatz after title in par headers"),
            );

            let value = absatz.content;

            // We want ot use this information in our markdown output.
            // TODO: Use all metadata, instead of this specific call
            if key == "Beachte" {
                builder.add_next_para_note(value.clone());
            }

            self.metadata.insert(key, value);
            continue;
        }
    }

    // At the beginning of each 'Abschnitt' there are 4 static fields. Since they don't provide any
    // value, we skip them with this function.
    // If they are not there, we panic (unwrap), as we should take a look why they changed that.
    fn skip_static_fields(node: &mut Peekable<Children>) {
        Kzinhalt::parse(node.next().unwrap()); // "Bundesrecht konsolidiert"
        Kzinhalt::parse(node.next().unwrap()); // "Bundesrecht konsolidiert"
        Fzinhalt::parse(node.next().unwrap()); // "www.ris.bka.gv.at" and "Seite X von Y"
        Fzinhalt::parse(node.next().unwrap()); // "www.ris.bka.gv.at" and "Seite X von Y"
    }

    // we have optionally headers. Such as "Einleitung", "Von den bürgerlichen Gesetzen üerhaupt,"
    // etc. If we have headers which indicate that we are done and we want to stop parsing
    // ("anlage" + "Artikel" we indicate this wish by returning false.
    fn handle_headers(&self, c: &mut Peekable<Children>, builder: &mut LawBuilder) -> bool {
        while let Some(child) = c.peek() {
            // Schiffahrtsgesetz: stop @ anlagen (for now)
            if Ueberschrift::test(child, "anlage") {
                return false;
            }
            if Ueberschrift::test(child, "g1") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g1");
                if ueberschrift.content.trim().starts_with("Artikel") {
                    return false;
                }
                builder.new_header(&ueberschrift.content);
            } else if Ueberschrift::test(child, "g2") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g2");
                builder.new_desc(&ueberschrift.content);
            } else if Ueberschrift::test(child, "g1min") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "g1min");
                builder.new_header(&ueberschrift.content);
            } else if Ueberschrift::test(child, "art") {
                let ueberschrift = Ueberschrift::parse(c.next().unwrap(), "art");
                if ueberschrift.content.trim().starts_with("Artikel") {
                    return false;
                }
            } else {
                break;
            }
        }

        // while (not if) because we can have two subsequent paraheaders (e.g. § 405 abgb)
        while let Some(child) = c.peek() {
            if Ueberschrift::test(child, "para") {
                builder
                    .new_next_para_header(&Ueberschrift::parse(c.next().unwrap(), "para").content);
                continue;
            }
            break;
        }
        true
    }
}
